@@ -20,6 +20,7 @@
 
 #define PY_SSIZE_T_CLEAN
 #include "Python.h"
+#include "thirdparty/pythoncapi_compat.h"
 #include "libImaging/Imaging.h"
 
 #include <ft2build.h>
@@ -125,7 +126,8 @@ getfont(PyObject *self_, PyObject *args, PyObject *kw) {
     unsigned char *font_bytes;
     Py_ssize_t font_bytes_size = 0;
     static char *kwlist[] = {
-        "filename", "size", "index", "encoding", "font_bytes", "layout_engine", NULL};
+        "filename", "size", "index", "encoding", "font_bytes", "layout_engine", NULL
+    };
 
     if (!library) {
         PyErr_SetString(PyExc_OSError, "failed to initialize FreeType library");
@@ -147,7 +149,8 @@ getfont(PyObject *self_, PyObject *args, PyObject *kw) {
             &encoding,
             &font_bytes,
             &font_bytes_size,
-            &layout_engine)) {
+            &layout_engine
+        )) {
         PyConfig_Clear(&config);
         return NULL;
     }
@@ -165,7 +168,8 @@ getfont(PyObject *self_, PyObject *args, PyObject *kw) {
             &encoding,
             &font_bytes,
             &font_bytes_size,
-            &layout_engine)) {
+            &layout_engine
+        )) {
         return NULL;
     }
 #endif
@@ -198,7 +202,8 @@ getfont(PyObject *self_, PyObject *args, PyObject *kw) {
                 (FT_Byte *)self->font_bytes,
                 font_bytes_size,
                 index,
-                &self->face);
+                &self->face
+            );
         }
     }
 
@@ -242,7 +247,8 @@ text_layout_raqm(
     const char *dir,
     PyObject *features,
     const char *lang,
-    GlyphInfo **glyph_info) {
+    GlyphInfo **glyph_info
+) {
     size_t i = 0, count = 0, start = 0;
     raqm_t *rq;
     raqm_glyph_t *glyphs = NULL;
@@ -266,7 +272,7 @@ text_layout_raqm(
         }
         set_text = raqm_set_text(rq, text, size);
         PyMem_Free(text);
-    } else {
+    } else if (PyBytes_Check(string)) {
         char *buffer;
         PyBytes_AsStringAndSize(string, &buffer, &size);
         if (!buffer || !size) {
@@ -275,6 +281,9 @@ text_layout_raqm(
             goto failed;
         }
         set_text = raqm_set_text_utf8(rq, buffer, size);
+    } else {
+        PyErr_SetString(PyExc_TypeError, "expected string or bytes");
+        goto failed;
     }
     if (!set_text) {
         PyErr_SetString(PyExc_ValueError, "raqm_set_text() failed");
@@ -296,13 +305,14 @@ text_layout_raqm(
 #if !defined(RAQM_VERSION_ATLEAST)
             /* RAQM_VERSION_ATLEAST was added in Raqm 0.7.0 */
             PyErr_SetString(
-                PyExc_ValueError,
-                "libraqm 0.7 or greater required for 'ttb' direction");
+                PyExc_ValueError, "libraqm 0.7 or greater required for 'ttb' direction"
+            );
             goto failed;
 #endif
         } else {
             PyErr_SetString(
-                PyExc_ValueError, "direction must be either 'rtl', 'ltr' or 'ttb'");
+                PyExc_ValueError, "direction must be either 'rtl', 'ltr' or 'ttb'"
+            );
             goto failed;
         }
     }
@@ -398,7 +408,8 @@ text_layout_fallback(
     const char *lang,
     GlyphInfo **glyph_info,
     int mask,
-    int color) {
+    int color
+) {
     int error, load_flags, i;
     char *buffer = NULL;
     FT_ULong ch;
@@ -411,13 +422,17 @@ text_layout_fallback(
         PyErr_SetString(
             PyExc_KeyError,
             "setting text direction, language or font features is not supported "
-            "without libraqm");
+            "without libraqm"
+        );
     }
 
     if (PyUnicode_Check(string)) {
         count = PyUnicode_GET_LENGTH(string);
-    } else {
+    } else if (PyBytes_Check(string)) {
         PyBytes_AsStringAndSize(string, &buffer, &count);
+    } else {
+        PyErr_SetString(PyExc_TypeError, "expected string or bytes");
+        return 0;
     }
     if (count == 0) {
         return 0;
@@ -458,7 +473,8 @@ text_layout_fallback(
                     last_index,
                     (*glyph_info)[i].index,
                     ft_kerning_default,
-                    &delta) == 0) {
+                    &delta
+                ) == 0) {
                 (*glyph_info)[i - 1].x_advance += PIXEL(delta.x);
                 (*glyph_info)[i - 1].y_advance += PIXEL(delta.y);
             }
@@ -482,7 +498,8 @@ text_layout(
     const char *lang,
     GlyphInfo **glyph_info,
     int mask,
-    int color) {
+    int color
+) {
     size_t count;
 #ifdef HAVE_RAQM
     if (have_raqm && self->layout_engine == LAYOUT_RAQM) {
@@ -491,7 +508,8 @@ text_layout(
 #endif
     {
         count = text_layout_fallback(
-            string, self, dir, features, lang, glyph_info, mask, color);
+            string, self, dir, features, lang, glyph_info, mask, color
+        );
     }
     return count;
 }
@@ -513,7 +531,8 @@ font_getlength(FontObject *self, PyObject *args) {
     /* calculate size and bearing for a given string */
 
     if (!PyArg_ParseTuple(
-            args, "O|zzOz:getlength", &string, &mode, &dir, &features, &lang)) {
+            args, "O|zzOz:getlength", &string, &mode, &dir, &features, &lang
+        )) {
         return NULL;
     }
 
@@ -555,7 +574,8 @@ bounding_box_and_anchors(
     int *width,
     int *height,
     int *x_offset,
-    int *y_offset) {
+    int *y_offset
+) {
     int position; /* pen position along primary axis, in 26.6 precision */
     int advanced; /* pen position along primary axis, in pixels */
     int px, py;   /* position of current glyph, in pixels */
@@ -660,7 +680,8 @@ bounding_box_and_anchors(
                 case 'm':  // middle (ascender + descender) / 2
                     y_anchor = PIXEL(
                         (face->size->metrics.ascender + face->size->metrics.descender) /
-                        2);
+                        2
+                    );
                     break;
                 case 's':  // horizontal baseline
                     y_anchor = 0;
@@ -740,7 +761,8 @@ font_getsize(FontObject *self, PyObject *args) {
     /* calculate size and bearing for a given string */
 
     if (!PyArg_ParseTuple(
-            args, "O|zzOzz:getsize", &string, &mode, &dir, &features, &lang, &anchor)) {
+            args, "O|zzOzz:getsize", &string, &mode, &dir, &features, &lang, &anchor
+        )) {
         return NULL;
     }
 
@@ -772,7 +794,8 @@ font_getsize(FontObject *self, PyObject *args) {
         &width,
         &height,
         &x_offset,
-        &y_offset);
+        &y_offset
+    );
     if (glyph_info) {
         PyMem_Free(glyph_info);
         glyph_info = NULL;
@@ -807,10 +830,9 @@ font_render(FontObject *self, PyObject *args) {
     unsigned char convert_scale;    /* scale factor for non-8bpp bitmaps */
     PyObject *image;
     Imaging im;
-    Py_ssize_t id;
     int mask = 0;  /* is FT_LOAD_TARGET_MONO enabled? */
     int color = 0; /* is FT_LOAD_COLOR enabled? */
-    int stroke_width = 0;
+    float stroke_width = 0;
     PY_LONG_LONG foreground_ink_long = 0;
     unsigned int foreground_ink;
     const char *mode = NULL;
@@ -830,7 +852,7 @@ font_render(FontObject *self, PyObject *args) {
 
     if (!PyArg_ParseTuple(
             args,
-            "OO|zzOzizLffO:render",
+            "OO|zzOzfzLffO:render",
             &string,
             &fill,
             &mode,
@@ -841,7 +863,8 @@ font_render(FontObject *self, PyObject *args) {
             &anchor,
             &foreground_ink_long,
             &x_start,
-            &y_start)) {
+            &y_start
+        )) {
         return NULL;
     }
 
@@ -888,29 +911,26 @@ font_render(FontObject *self, PyObject *args) {
         &width,
         &height,
         &x_offset,
-        &y_offset);
+        &y_offset
+    );
     if (error) {
         PyMem_Del(glyph_info);
         return NULL;
     }
 
-    width += stroke_width * 2 + ceil(x_start);
-    height += stroke_width * 2 + ceil(y_start);
+    width += ceil(stroke_width * 2 + x_start);
+    height += ceil(stroke_width * 2 + y_start);
     image = PyObject_CallFunction(fill, "ii", width, height);
-    if (image == Py_None) {
-        PyMem_Del(glyph_info);
-        return Py_BuildValue("N(ii)", image, 0, 0);
-    } else if (image == NULL) {
+    if (image == NULL) {
         PyMem_Del(glyph_info);
         return NULL;
     }
-    PyObject *imageId = PyObject_GetAttrString(image, "id");
-    id = PyLong_AsSsize_t(imageId);
-    Py_XDECREF(imageId);
-    im = (Imaging)id;
+    PyObject *imagePtr = PyObject_GetAttrString(image, "ptr");
+    im = (Imaging)PyCapsule_GetPointer(imagePtr, IMAGING_MAGIC);
+    Py_XDECREF(imagePtr);
 
-    x_offset -= stroke_width;
-    y_offset -= stroke_width;
+    x_offset = round(x_offset - stroke_width);
+    y_offset = round(y_offset - stroke_width);
     if (count == 0 || width == 0 || height == 0) {
         PyMem_Del(glyph_info);
         return Py_BuildValue("N(ii)", image, x_offset, y_offset);
@@ -925,10 +945,11 @@ font_render(FontObject *self, PyObject *args) {
 
         FT_Stroker_Set(
             stroker,
-            (FT_Fixed)stroke_width * 64,
+            (FT_Fixed)round(stroke_width * 64),
             FT_STROKER_LINECAP_ROUND,
             FT_STROKER_LINEJOIN_ROUND,
-            0);
+            0
+        );
     }
 
     /*
@@ -962,8 +983,8 @@ font_render(FontObject *self, PyObject *args) {
     }
 
     /* set pen position to text origin */
-    x = (-x_min + stroke_width + x_start) * 64;
-    y = (-y_max + (-stroke_width) - y_start) * 64;
+    x = round((-x_min + stroke_width + x_start) * 64);
+    y = round((-y_max + (-stroke_width) - y_start) * 64);
 
     if (stroker == NULL) {
         load_flags |= FT_LOAD_RENDER;
@@ -1103,8 +1124,8 @@ font_render(FontObject *self, PyObject *args) {
                                     BLEND(src_alpha, target[k * 4 + 2], src_blue, tmp);
                                 target[k * 4 + 3] = CLIP8(
                                     src_alpha +
-                                    MULDIV255(
-                                        target[k * 4 + 3], (255 - src_alpha), tmp));
+                                    MULDIV255(target[k * 4 + 3], (255 - src_alpha), tmp)
+                                );
                             } else {
                                 /* paste unpremultiplied RGBA values */
                                 target[k * 4 + 0] = src_red;
@@ -1122,15 +1143,20 @@ font_render(FontObject *self, PyObject *args) {
                             if (src_alpha > 0) {
                                 if (target[k * 4 + 3] > 0) {
                                     target[k * 4 + 0] = BLEND(
-                                        src_alpha, target[k * 4 + 0], ink[0], tmp);
+                                        src_alpha, target[k * 4 + 0], ink[0], tmp
+                                    );
                                     target[k * 4 + 1] = BLEND(
-                                        src_alpha, target[k * 4 + 1], ink[1], tmp);
+                                        src_alpha, target[k * 4 + 1], ink[1], tmp
+                                    );
                                     target[k * 4 + 2] = BLEND(
-                                        src_alpha, target[k * 4 + 2], ink[2], tmp);
+                                        src_alpha, target[k * 4 + 2], ink[2], tmp
+                                    );
                                     target[k * 4 + 3] = CLIP8(
                                         src_alpha +
                                         MULDIV255(
-                                            target[k * 4 + 3], (255 - src_alpha), tmp));
+                                            target[k * 4 + 3], (255 - src_alpha), tmp
+                                        )
+                                    );
                                 } else {
                                     target[k * 4 + 0] = ink[0];
                                     target[k * 4 + 1] = ink[1];
@@ -1148,7 +1174,9 @@ font_render(FontObject *self, PyObject *args) {
                                         ? CLIP8(
                                               src_alpha +
                                               MULDIV255(
-                                                  target[k], (255 - src_alpha), tmp))
+                                                  target[k], (255 - src_alpha), tmp
+                                              )
+                                          )
                                         : src_alpha;
                             }
                         }
@@ -1209,30 +1237,49 @@ font_getvarnames(FontObject *self) {
         return NULL;
     }
 
+    int *list_names_filled = PyMem_Malloc(num_namedstyles * sizeof(int));
+    if (list_names_filled == NULL) {
+        Py_DECREF(list_names);
+        FT_Done_MM_Var(library, master);
+        return PyErr_NoMemory();
+    }
+
+    for (unsigned int i = 0; i < num_namedstyles; i++) {
+        list_names_filled[i] = 0;
+    }
+
     name_count = FT_Get_Sfnt_Name_Count(self->face);
     for (i = 0; i < name_count; i++) {
         error = FT_Get_Sfnt_Name(self->face, i, &name);
         if (error) {
+            PyMem_Free(list_names_filled);
             Py_DECREF(list_names);
             FT_Done_MM_Var(library, master);
             return geterror(error);
         }
 
         for (j = 0; j < num_namedstyles; j++) {
-            if (PyList_GetItem(list_names, j) != NULL) {
+            if (list_names_filled[j]) {
                 continue;
             }
 
             if (master->namedstyle[j].strid == name.name_id) {
                 list_name = Py_BuildValue("y#", name.string, name.string_len);
+                if (list_name == NULL) {
+                    PyMem_Free(list_names_filled);
+                    Py_DECREF(list_names);
+                    FT_Done_MM_Var(library, master);
+                    return NULL;
+                }
                 PyList_SetItem(list_names, j, list_name);
+                list_names_filled[j] = 1;
                 break;
             }
         }
     }
 
+    PyMem_Free(list_names_filled);
     FT_Done_MM_Var(library, master);
-
     return list_names;
 }
 
@@ -1289,9 +1336,14 @@ font_getvaraxes(FontObject *self) {
 
             if (name.name_id == axis.strid) {
                 axis_name = Py_BuildValue("y#", name.string, name.string_len);
-                PyDict_SetItemString(
-                    list_axis, "name", axis_name ? axis_name : Py_None);
-                Py_XDECREF(axis_name);
+                if (axis_name == NULL) {
+                    Py_DECREF(list_axis);
+                    Py_DECREF(list_axes);
+                    FT_Done_MM_Var(library, master);
+                    return NULL;
+                }
+                PyDict_SetItemString(list_axis, "name", axis_name);
+                Py_DECREF(axis_name);
                 break;
             }
         }
@@ -1345,7 +1397,12 @@ font_setvaraxes(FontObject *self, PyObject *args) {
         return PyErr_NoMemory();
     }
     for (i = 0; i < num_coords; i++) {
-        item = PyList_GET_ITEM(axes, i);
+        item = PyList_GetItemRef(axes, i);
+        if (item == NULL) {
+            free(coords);
+            return NULL;
+        }
+
         if (PyFloat_Check(item)) {
             coord = PyFloat_AS_DOUBLE(item);
         } else if (PyLong_Check(item)) {
@@ -1353,10 +1410,12 @@ font_setvaraxes(FontObject *self, PyObject *args) {
         } else if (PyNumber_Check(item)) {
             coord = PyFloat_AsDouble(item);
         } else {
+            Py_DECREF(item);
             free(coords);
             PyErr_SetString(PyExc_TypeError, "list must contain numbers");
             return NULL;
         }
+        Py_DECREF(item);
         coords[i] = coord * 65536;
     }
 
@@ -1393,7 +1452,8 @@ static PyMethodDef font_methods[] = {
     {"setvarname", (PyCFunction)font_setvarname, METH_VARARGS},
     {"setvaraxes", (PyCFunction)font_setvaraxes, METH_VARARGS},
 #endif
-    {NULL, NULL}};
+    {NULL, NULL}
+};
 
 static PyObject *
 font_getattr_family(FontObject *self, void *closure) {
@@ -1450,7 +1510,8 @@ static struct PyGetSetDef font_getsetters[] = {
     {"x_ppem", (getter)font_getattr_x_ppem},
     {"y_ppem", (getter)font_getattr_y_ppem},
     {"glyphs", (getter)font_getattr_glyphs},
-    {NULL}};
+    {NULL}
+};
 
 static PyTypeObject Font_Type = {
     PyVarObject_HEAD_INIT(NULL, 0) "Font", /*tp_name*/
@@ -1486,7 +1547,8 @@ static PyTypeObject Font_Type = {
 };
 
 static PyMethodDef _functions[] = {
-    {"getfont", (PyCFunction)getfont, METH_VARARGS | METH_KEYWORDS}, {NULL, NULL}};
+    {"getfont", (PyCFunction)getfont, METH_VARARGS | METH_KEYWORDS}, {NULL, NULL}
+};
 
 static int
 setup_module(PyObject *m) {
@@ -1576,5 +1638,9 @@ PyInit__imagingft(void) {
         return NULL;
     }
 
+#ifdef Py_GIL_DISABLED
+    PyUnstable_Module_SetGIL(m, Py_MOD_GIL_NOT_USED);
+#endif
+
     return m;
 }
