@@ -89,7 +89,8 @@ j2k_seek(OPJ_OFF_T p_nb_bytes, void *p_user_data) {
 /* -------------------------------------------------------------------- */
 
 typedef void (*j2k_pack_tile_t)(
-    Imaging im, UINT8 *buf, unsigned x0, unsigned y0, unsigned w, unsigned h);
+    Imaging im, UINT8 *buf, unsigned x0, unsigned y0, unsigned w, unsigned h
+);
 
 static void
 j2k_pack_l(Imaging im, UINT8 *buf, unsigned x0, unsigned y0, unsigned w, unsigned h) {
@@ -157,7 +158,8 @@ j2k_pack_rgb(Imaging im, UINT8 *buf, unsigned x0, unsigned y0, unsigned w, unsig
 
 static void
 j2k_pack_rgba(
-    Imaging im, UINT8 *buf, unsigned x0, unsigned y0, unsigned w, unsigned h) {
+    Imaging im, UINT8 *buf, unsigned x0, unsigned y0, unsigned w, unsigned h
+) {
     UINT8 *pr = buf;
     UINT8 *pg = pr + w * h;
     UINT8 *pb = pg + w * h;
@@ -205,8 +207,8 @@ j2k_set_cinema_params(Imaging im, int components, opj_cparameters_t *params) {
 
     if (params->cp_cinema == OPJ_CINEMA4K_24) {
         float max_rate =
-            ((float)(components * im->xsize * im->ysize * 8) /
-             (CINEMA_24_CS_LENGTH * 8));
+            ((float)(components * im->xsize * im->ysize * 8) / (CINEMA_24_CS_LENGTH * 8)
+            );
 
         params->POC[0].tile = 1;
         params->POC[0].resno0 = 0;
@@ -241,8 +243,8 @@ j2k_set_cinema_params(Imaging im, int components, opj_cparameters_t *params) {
         params->max_comp_size = COMP_24_CS_MAX_LENGTH;
     } else {
         float max_rate =
-            ((float)(components * im->xsize * im->ysize * 8) /
-             (CINEMA_48_CS_LENGTH * 8));
+            ((float)(components * im->xsize * im->ysize * 8) / (CINEMA_48_CS_LENGTH * 8)
+            );
 
         for (n = 0; n < params->tcp_numlayers; ++n) {
             rate = 0;
@@ -650,10 +652,3 @@ ImagingJpeg2KEncodeCleanup(ImagingCodecState state) {
 }
 
 #endif /* HAVE_OPENJPEG */
-
-/*
- * Local Variables:
- * c-basic-offset: 4
- * End:
- *
- */
